const { args } = require('./helpers')
const Lines = require('../lines')

const gs = (g) => (contexts, separator, last_separator) => {
  s = ''
  if (!separator) {
    separator = ' '
  }
  if (!last_separator) {
    last_separator = separator
  }
  next_separator = ''
  for (let i = 0; i < contexts.length; ++i) {
    const context = contexts[i]
    value = g(context)
    if (i > 0) {
      if (i == contexts.length - 1) {
        next_separator = last_separator
      } else {
        next_separator = separator
      }
    }
    s += next_separator + value
  }
  return s
}

class Generator {
  constructor ({ match, apply, uuid, index, km }) {
    this.match = match
    this._apply = apply
    this.uuid = uuid
    this.index = index
    this.km = km
  }

  getAPI (config) {
    if (config && config.getAPI) {
      return config.getAPI(this.uuid)
    }
  }

  toString () {
    return `Generator(${this.match}, ${this._apply})`
  }

  matches (objects, context, hierarchy, config) {
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    return this.match({ args: args(context, hierarchy), objects, global: objects, hierarchy, context, api: this.getAPI(config) })
  }

  apply (objects, g, context, hierarchy, config, response) {
    this.getAPI(config)
    let n = (id) => id
    if (config && 'nsToString' in config) {
      n = (id) => config.nsToString(id)
    }
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    return this._apply({ args: args(context, hierarchy), objects, global: objects, g, n, hierarchy, context, uuid: this.uuid, gs: gs(g), config, response, api: this.getAPI(config) })
  }
}

class Generators {
  constructor (generators) {
    this.generators = generators
    this.logs = []
  };

  // assumed - properties added to contexts before the generators are called. For setting paraphrase property
  apply (objects, contexts, hierarchy, config, response, assumed = {}) {
    if (Array.isArray(contexts)) {
    } else if (typeof contexts === 'object') {
      contexts = [contexts]
    } else {
      return String(contexts)
    }

    const contexts_prime = []
    for (const icontext in contexts) {
      let context = contexts[icontext]
      context = Object.assign({}, context, assumed)
      // let context_prime = JSON.stringify(context);
      let context_prime = context
      for (let igenerator = 0; igenerator < this.generators.length; ++igenerator) {
        const generator = this.generators[igenerator]
        if (generator.matches(objects, context, hierarchy, config)) {
          const g = (context) => this.apply(objects, context, hierarchy, config, response, assumed)
          // this.logs.push(`Generators: applied ${generator.toString()}\n to\n ${JSON.stringify(context)}`)
          context_prime = generator.apply(objects, g, context, hierarchy, config, response)
          {
            const widths = [10, 10, 90]
            const lines = new Lines(widths)
            lines.setElement(0, 0, 'Generator')
            if (generator.index > -1 && generator.km) {
              lines.setElement(0, 2, `KM '${generator.km}' ordinal: ${generator.index}`)
            }
            lines.newRow()
            lines.setElement(0, 1, 'APPLIED')
            lines.setElement(0, 2, generator.toString())
            lines.newRow()
            lines.setElement(0, 1, 'TO')
            lines.setElement(0, 2, JSON.stringify(context, null, 2))
            lines.newRow()
            lines.setElement(0, 1, 'RESULT')
            lines.setElement(0, 2, context_prime)
            this.logs.push(lines.toString())
          }
          break
        }
      }
      contexts_prime.push(context_prime)
    }
    return contexts_prime
  }
}

module.exports = { Generator, Generators }
